package main

// Simulate the function of a bluetooth enabled geiger counter. Responds correctly to input, and transmitts
// random data corresponding to measured counts.

import (
	"time"
	"log"
	"github.com/tarm/serial"
	"math/rand"
	"strconv"
	"fmt"
	"bytes"
	"os"
	"bufio"
	"math"
)


const (
	// Radioactive source simulation
	RADIOACTIVITY_LEVEL = 400.0			// What the most intense count should be
	VARIATION_FREQ = 0.0002				// Movement frequency
	RADIO_VARIATION = 20				// How many counts to vary by

	// GM Counter specifics
	TIME_INTERVAL = 800					// Time between count refreshes [ms]
	GM_TUBE_MAX_COUNT = 2000			// Maximum counting rate [s^-1]
	GM_TUBE_NAME = "SBM-19\n"			// String literal of name
	GM_TUBE_DOSE_RATE = "875.1\n"		// CPM / uSv/h conversion factor
	
	// To GM counter
	REQUEST_GM_TUBE_DATA = "READC"
	START_GM = "START"
	STOP_GM = "HALTT"

	// To app
	HEADER_TIME = "PERID:"
	HEADER_MAXCOUNT = "MAXCT:"
	HEADER_NAME = "NAMET:"
	HEADER_DOSERATE = "DOSER:"
	DATAGRAM_COUNT = "COUNT:"

)


// Variable used to control sending of Geiger data
var enableTransmitt bool
// Global radioactive level variable
var radioactivity float64
var timeTracker float64

// This function creates a random GM tube reading, and transmitts it
func createGeigerEvent(comPort *serial.Port) (success bool) {
	
	// Measure the current radioactivity level and transmit it
	gmCounts := int(radioactivity);
	
	// Format it into a byte sequence
	dataString := strconv.Itoa(gmCounts)
	dataString = DATAGRAM_COUNT + dataString + "\n";
	
	
	// Send the data
	if enableTransmitt {
		// Write some bytes to the port
		_, err := comPort.Write([]byte(dataString))
		if err != nil {
			log.Fatal(err)
			return false;
		}
	}
	
	return true;
}

// Read a line of data from the serial 
func readLineFromSerial(comPort *serial.Port) (stringIndex int, stringBuffer []byte) {
	buf := make([]byte, 128)
	lfReceived := false;
	var n int
	var err error
	
	stringBuffer = make([]byte, 256)
	stringIndex = 0;
	
	// Read from the serial port until reaching a LF character
	for !lfReceived {
		// Read some bytes from the port
		n, err = comPort.Read(buf)
		
		// Check if it went alright
		if err != nil {
			log.Fatal(err)
			
		}else {
			
			// Check what we got
			if n > 0 {
			
			// Print buffer contents
			//fmt.Println("Buffer contents: ", string(buf[:n]))
			//fmt.Println("String index: ", stringIndex)
			
				// Loop through all received bytes
				for i:=0; i<n; i++ {
				
					if (buf[i] == '\n') {
						// Terminate the string, and stop putting received chars in the string buffer
						stringBuffer[stringIndex] = buf[i];
						stringBuffer[stringIndex+1] = 0;
						lfReceived = true;
						break;
						
					}else {
						// Store the received char
						stringBuffer[stringIndex] = buf[i];
					}
					
					// Move the string index up
					stringIndex++;
					
				}
				
			}
			
		}
		
	}
	
	
	return
}


// Parse a string of data
func parseDataString(inputString []byte, comPort *serial.Port) {

	getCode := []byte(REQUEST_GM_TUBE_DATA);
	startCode := []byte(START_GM);
	stopCode := []byte(STOP_GM);

	if (bytes.Compare(inputString, startCode) == 0) {
		// Enable serial transmission
		enableTransmitt = true;
		
	}else if (bytes.Compare(inputString, stopCode) == 0) {
		// Disable serial transmission
		enableTransmitt = false;
		
	}else if (bytes.Compare(inputString, getCode) == 0) {
		// Send all relevant GM tube specs
		
		// Format the data
		intervalString := HEADER_TIME + strconv.Itoa(TIME_INTERVAL) + "\n";
		maxcountString := HEADER_MAXCOUNT + strconv.Itoa(GM_TUBE_MAX_COUNT) + "\n";
		nameString := HEADER_NAME + GM_TUBE_NAME + "\n";
		doseString := HEADER_DOSERATE + GM_TUBE_DOSE_RATE + "\n";
		
		// Send all data
		dataString := intervalString + maxcountString + nameString + doseString;
		_, err := comPort.Write([]byte(dataString))
		if err != nil {
			log.Fatal(err)
		}
		
	}
}

// Main takes single argument, which is COM port to use.
func main() {


	// Check command line args
	valid := false;
	portName := "COM1";
	
	if (len(os.Args) > 1) {
		// We've been passed something, parse it
		arg1 := []byte(os.Args[1]);
		
		// Check if it corresponds to a port in the range 1-99
		for i := 0; i < 100; i++ {
			currentPort := []byte("COM"+ strconv.Itoa(i));
			if (bytes.Compare(arg1, currentPort) == 0) {
				valid = true;
				fmt.Println("Connecting to port: ", string(arg1));
				portName = string(arg1);
				break;
			}
		}
	}
	
	if (!valid) {
		log.Fatal("Invalid argument. Use COMx, where x is a port number between 1 - 99. E.g. COM4. Baud rate is 9600.")
	}

	// Initial state is to not send data over serial line
	enableTransmitt = false;

	// Connect to the serial port
	c := &serial.Config{Name: portName, Baud: 9600}
	s, err := serial.OpenPort(c)
	if err != nil {
		log.Fatal(err)
	}
	
	
	// Timer, used to simulate the movement of a radioactive check source
	timeBase := time.NewTicker(1 * time.Millisecond)
	quit := make(chan struct{})
	go func() {
		for {
		   select {
			case <- timeBase.C:
				// Update the current radioactivity level
				timeTracker++;
				radioactivity = RADIOACTIVITY_LEVEL*(1.0 + math.Sin(2*math.Pi*VARIATION_FREQ*timeTracker));
				radioactivity = radioactivity + float64(rand.Intn(RADIO_VARIATION));
				
			case <- quit:
				timeBase.Stop()
				return
			}
		}
	}()
	
	
	
	// Timer, so geiger counts are sent at a set interval
	ticker := time.NewTicker(TIME_INTERVAL * time.Millisecond)
	go func() {
		for {
		   select {
			case <- ticker.C:
				// Create a random count value, and send it
				go createGeigerEvent(s);
			case <- quit:
				ticker.Stop()
				return
			}
		}
	}()
	
	
	
	// Create a thread which reads from serial, and responds correctly
	go func() {
		for {
		   select {
			case <- quit:
				return
			default:
				// Read from serial port
				readBytes, data := readLineFromSerial(s);
				
				// Check if any data was received
				if (readBytes != 0) {
					// Parse it
					fmt.Println("Recv data: ", string(data[:readBytes]))
					
					parseDataString(data[:readBytes], s);
				}
			}
		}
	}()
	
	
	// Exit on keypress
	fmt.Println("Press enter to quit...")
	reader := bufio.NewReader(os.Stdin)
    reader.ReadString('\n')
	
	
	// Shut down go routines
	close(quit);
}